extends Node3D

@export var tracking : UltraleapHandTracking
@export var materials_to_rectify : Array[Material]

@export_group("Layers")
@export var disconnected : CanvasLayer
@export var no_device : CanvasLayer

signal presence_of_device_changed(is_present : bool)
signal current_device_changed(device : UltraleapDevice)

var previous_tracking_mode : int


func _on_device_selection(serial):
	var dev : UltraleapDevice = _get_device_from_serial(serial)
	var current_dev : UltraleapDevice = tracking.get_primary_device()
	
	if current_dev == null or current_dev.id != dev.id:
		tracking.set_primary_device(dev)
		emit_signal("current_device_changed", dev)


func _get_device_from_serial(serial : String):
	for device in tracking.devices_list:
		if device.serial == serial:
			return device


func _on_hand_tracking_device_added(device):
	if tracking.devices_list.size() > 0:
		emit_signal("presence_of_device_changed", true)
		no_device.hide()


func _on_hand_tracking_device_removed(_device : UltraleapDevice):
	if tracking.devices_list.size() == 0:
		emit_signal("presence_of_device_changed", false)
		no_device.show()
	if tracking.devices_primary_device == null:
		current_device_changed.emit(null)


func on_undistort_toggled(enabled):
	for material in materials_to_rectify:
		material.set_shader_parameter("undistort", enabled)


func _on_connection_status_changed(connected):
	if connected:
		emit_signal("presence_of_device_changed", false)
		no_device.show()
		disconnected.hide()
	else:
		emit_signal("presence_of_device_changed", false)
		no_device.hide()
		disconnected.show()
